/*	  Copyright (C) 2000,2001,2002  Sony Computer Entertainment America
       	  
       	  This file is subject to the terms and conditions of the GNU Lesser
	  General Public License Version 2.1. See the file "COPYING" in the
	  main directory of this archive for more details.                             */

/********************************************
 * Caution!
 *
 * This was written as a test/example.. It is NOT
 * an efficient implementation.
 */

#include "cloth.h"
#include "vu1_format.h"

#include "ps2s/timer.h"
#include "ps2s/utils.h"
#include "ps2s/matrix.h"

cloth::cloth( int _num_rows, int _num_cols,
	      float _side_len_row, float _side_len_col,
	      vec_xyzw _lower_left_corner )
   : gravity( 0, -0.2f, 0, 0 ),
     num_rows(_num_rows), num_cols(_num_cols),
     side_len_row(_side_len_row), side_len_col(_side_len_col),
     lower_left_corner(_lower_left_corner),
     stretch_rest_length_row(side_len_row/(float)num_cols),
     stretch_rest_length_col(side_len_col/(float)num_rows),
     shear_rest_length(sqrtf(stretch_rest_length_row*stretch_rest_length_row
			     + stretch_rest_length_col*stretch_rest_length_col)),
     bend_rest_length_row(stretch_rest_length_row * 2.0f),
     bend_rest_length_col(stretch_rest_length_col * 2.0f)
{
   points = new cloth_pt[num_cols * num_rows];
   reset();
}

cloth::~cloth( )
{
   delete [] points;
}

void
cloth::reset( )
{
   vec_xyzw upper_left = lower_left_corner;
   upper_left = (vec_x)upper_left - vec_x( side_len_col/2.0f );
   vec_xyzw col_inc( 0, stretch_rest_length_row, 0, 0 ),
      row_inc( stretch_rest_length_col, 0, 0, 0 );
   vec_xyzw cur_pos = upper_left, cur_row_start = upper_left;
   vec_xyzw zero(0, 0, 0, 0);
   cloth_pt *cur_pt = points;
   int col, row;
   for ( row = 0; row < num_rows; row++ ) {
      for ( col = 0; col < num_cols; col++ ) {
	 cur_pt->position = cur_pos;
	 cur_pt->velocity = zero;

	 cur_pt++;
	 cur_pos += col_inc;
      }
      cur_row_start += row_inc;
      cur_pos = cur_row_start;
   }

   update_normals( points );
}

vec_xyzw
cloth::get_spring_force( const cloth_pt &pt1, const cloth_pt &pt2,
			 float spring_const, float damp_const, float rest_length )
{
   vec_xyzw delta_pos = pt2.position - pt1.position;
   float distance = delta_pos.length();
   float hooke_term = spring_const * (distance - rest_length);

   vec_xyzw delta_vel = pt2.velocity - pt1.velocity;
   float damp_term = damp_const * (float)delta_vel.dot(delta_pos) / distance;

   vec_xyzw force = delta_pos.normalize();

   force *= hooke_term + damp_term;

   return force;
}

void
cloth::update( float timestep, vec_xyzw wind_dir, float wind_strength ) {
   mStartTimer("cloth::update");

   cloth_pt *old_pts = points, *new_pts = ((cloth_pt*)Core::MemMappings::SP);

   int col, row;
   cloth_pt *cur_pt = old_pts, *new_pt = new_pts;

   cloth_pt *n_pt, *s_pt, *e_pt, *w_pt,
      *nw_pt, *ne_pt, *sw_pt, *se_pt,
      *n_n_pt, *s_s_pt, *e_e_pt, *w_w_pt;
   for ( row = 0; row < num_rows; row++ ) {
      for ( col = 0; col < num_cols; col++ ) {
	 w_pt = cur_pt - 1;
	 e_pt = cur_pt + 1;
	 n_pt = cur_pt - num_cols;
	 s_pt = cur_pt + num_cols;

	 w_w_pt = w_pt - 1;
	 e_e_pt = e_pt + 1;
	 n_n_pt = n_pt - num_cols;
	 s_s_pt = s_pt + num_cols;

	 nw_pt = n_pt - 1;
	 ne_pt = n_pt + 1;
	 sw_pt = s_pt - 1;
	 se_pt = s_pt + 1;

	 vec_xyzw up(0, 0, 0, 0), down, left, right;
	 down = left = right = up;

	 // stretch forces

	 float spring_const = 4.0f;
	 float damp_const = 1.5f;
	 float rest_length_row = stretch_rest_length_row,
	    rest_length_col = stretch_rest_length_col;

	 if ( row != 0 ) 
	    up = get_spring_force( *cur_pt, *n_pt, spring_const, damp_const, rest_length_col );
	 if ( row != (num_rows-1) ) 
	    down = get_spring_force( *cur_pt, *s_pt, spring_const, damp_const, rest_length_col );
	 if ( col != 0 )
	    left = get_spring_force( *cur_pt, *w_pt, spring_const, damp_const, rest_length_row );
	 if ( col != (num_cols-1) )
	    right = get_spring_force( *cur_pt, *e_pt, spring_const, damp_const, rest_length_row );

	 // shear forces

	 spring_const = 4.0f;
	 damp_const = 1.5f;
	 float rest_length = shear_rest_length;

	 vec_xyzw top_bslash(0, 0, 0, 0), top_fslash, bottom_fslash, bottom_bslash;
	 top_fslash = bottom_bslash = bottom_fslash = top_bslash;

	 if ( row != 0 ) {
	    if ( col != 0 )
	       top_bslash = get_spring_force( *cur_pt, *nw_pt, spring_const, damp_const, rest_length );
	    if ( col != (num_cols-1) )
	       top_fslash = get_spring_force( *cur_pt, *ne_pt, spring_const, damp_const, rest_length );
	 }
	 if ( row != (num_rows-1) ) {
	    if ( col != 0 )
	       bottom_fslash = get_spring_force( *cur_pt, *sw_pt, spring_const, damp_const, rest_length );
	    if ( col != (num_cols-1) )
	       bottom_bslash = get_spring_force( *cur_pt, *se_pt, spring_const, damp_const, rest_length );
	 }

	 // bend forces

	 spring_const = 3.0f;
	 damp_const = 1.4f;
	 rest_length_row = bend_rest_length_row;
	 rest_length_col = bend_rest_length_col;

	 vec_xyzw n_bend(0, 0, 0, 0), s_bend, e_bend, w_bend;
	 s_bend = e_bend = w_bend = n_bend;

	 if ( row > 1 )
	    n_bend = get_spring_force( *cur_pt, *n_n_pt, spring_const, damp_const, rest_length_col );
	 if ( row < (num_rows-2) )
	    s_bend = get_spring_force( *cur_pt, *s_s_pt, spring_const, damp_const, rest_length_col );
	 if ( col > 1 )
	    w_bend = get_spring_force( *cur_pt, *w_w_pt, spring_const, damp_const, rest_length_row );
	 if ( col < (num_cols-2) )
	    e_bend = get_spring_force( *cur_pt, *e_e_pt, spring_const, damp_const, rest_length_row );

	 vec_xyzw force = up;
	 force += down;
	 force += left;
	 force += right;

	 force += top_fslash;
	 force += top_bslash;
	 force += bottom_bslash;
	 force += bottom_fslash;

	 force += n_bend;
	 force += s_bend;
	 force += e_bend;
	 force += w_bend;

	 // atmospheric drag

	 force += cur_pt->velocity * -0.3f;

	 // wind

	 vec_xyzw force_dir = cur_pt->normal;
	 force += force_dir * (cur_pt->normal.dot(wind_dir) * wind_strength);

	 // gravity

	 force += gravity;
					
	 vec_xyzw accel = force; // assuming mass of 1
	 vec_xyzw delta_v = accel * timestep;
	 vec_xyzw delta_pos = cur_pt->velocity * timestep;

	 // now fixup the lengths of the springs so that they (mostly) don't exceed
	 // their rest lengths by 10%.  Seems ok to only do this for the stretch
	 // springs

	 float length;
	 vec_xyzw direction;
			
	 vec_xyzw new_pos = cur_pt->position + delta_pos;

	 if ( row != 0 ) {
	    direction = n_pt->position - new_pos;
	    length = direction.length();
	    if ( length > stretch_rest_length_col * 1.1f ) {
	       direction /= length;
	       new_pos += direction * (length - stretch_rest_length_col * 1.1f);
	    }
	    else if ( length < stretch_rest_length_col * 0.9f ) {
	       direction /= length;
	       new_pos += direction * (length - stretch_rest_length_col * 0.9f);
	    }
	 }

	 if ( row != (num_rows-1) ) {
	    direction = s_pt->position - new_pos;
	    length = direction.length();
	    if ( length > stretch_rest_length_col * 1.1f ) {
	       direction /= length;
	       new_pos += direction * (length - stretch_rest_length_col * 1.1f);
	    }
	    else if ( length < stretch_rest_length_col * 0.9f ) {
	       direction /= length;
	       new_pos += direction * (length - stretch_rest_length_col * 0.9f);
	    }
	 }

	 if ( col != (num_cols-1) ) {
	    direction = e_pt->position - new_pos;
	    length = direction.length();
	    if ( length > stretch_rest_length_row * 1.1f ) {
	       direction /= length;
	       new_pos += direction * (length - stretch_rest_length_row * 1.1f);
	    }
	    else if ( length < stretch_rest_length_row * 0.9f ) {
	       direction /= length;
	       new_pos += direction * (length - stretch_rest_length_row * 0.9f);
	    }
	 }

	 if ( col != 0 ) {
	    direction = w_pt->position - new_pos;
	    length = direction.length();
	    if ( length > stretch_rest_length_row * 1.1f ) {
	       direction /= length;
	       new_pos += direction * (length - stretch_rest_length_row * 1.1f);
	    }
	    else if ( length < stretch_rest_length_row * 0.9f ) {
	       direction /= length;
	       new_pos += direction * (length - stretch_rest_length_row * 0.9f);
	    }
	 }

	 // find the velocity that would have resulted in the new final position
	 delta_pos = new_pos - cur_pt->position;
	 vec_xyzw new_vel = delta_pos / timestep;

	 // fix the leftmost col in place
	 if ( col == 0 ) {
	    *new_pt = *cur_pt;
	 }
	 else {
	    new_pt->position = new_pos;
	    // this has a nice stabilizing effect on the system.. I think the bend
	    // springs are losing it when the fabric becomes folded...
	    cur_pt->velocity = new_vel + delta_v;
	    new_pt->velocity = new_vel + delta_v;
	 }

	 new_pt++;
	 cur_pt++;
      }
   }

   update_normals( new_pts );

   // copy new points back
   Utils::MemCpy128( (tU128*)points, (tU128*)new_pts, num_cols*num_rows * (sizeof(cloth_pt)/16) );

   mStopTimer("cloth::update");
}


/********************************************
 * libvu0 vectors
 */

void
cloth::get_spring_force_lv( sceVu0FVECTOR result,
			    const cloth_pt_libvu0 &pt1, const cloth_pt_libvu0 &pt2,
			    float spring_const, float damp_const, float rest_length )
{
   sceVu0FVECTOR delta_pos;
   sceVu0SubVector( delta_pos,
		    const_cast<float*>(pt2.position), const_cast<float*>(pt1.position) );
   float distance = sqrtf( sceVu0InnerProduct(delta_pos, delta_pos) );
   /*
   vec_xyzw delta_pos = pt2.position - pt1.position;
   float distance = delta_pos.length();
   */


   float hooke_term = spring_const * (distance - rest_length);


   sceVu0FVECTOR delta_vel;
   sceVu0SubVector( delta_vel,
		    const_cast<float*>(pt2.velocity), const_cast<float*>(pt1.velocity) );
   float damp_term = damp_const * sceVu0InnerProduct( delta_vel, delta_pos );
   damp_term /= distance;
   /*
   vec_xyzw delta_vel = pt2.velocity - pt1.velocity;
   float damp_term = damp_const * (float)delta_vel.dot(delta_pos) / distance;
   */

   sceVu0FVECTOR force;
   sceVu0Normalize( force, delta_pos );
   /*
   vec_xyzw force = delta_pos.normalize();
   */

   sceVu0ScaleVector( result, force, hooke_term + damp_term );
   /*
   force *= hooke_term + damp_term;
   return force;
   */
}

void
cloth::update_libvu0( float timestep, sceVu0FVECTOR wind_dir, float wind_strength ) {
   mStartTimer("cloth::update_libvu0");

   cloth_pt_libvu0 *old_pts = (cloth_pt_libvu0*)points,
      *new_pts = ((cloth_pt_libvu0*)Core::MemMappings::SP);

   int col, row;
   cloth_pt_libvu0 *cur_pt = old_pts, *new_pt = new_pts;

   sceVu0FVECTOR gravity_lv;
   memcpy( (char*)gravity_lv, (char*)&gravity, 16 );

   cloth_pt_libvu0 *n_pt, *s_pt, *e_pt, *w_pt,
      *nw_pt, *ne_pt, *sw_pt, *se_pt,
      *n_n_pt, *s_s_pt, *e_e_pt, *w_w_pt;
   for ( row = 0; row < num_rows; row++ ) {
      for ( col = 0; col < num_cols; col++ ) {
	 w_pt = cur_pt - 1;
	 e_pt = cur_pt + 1;
	 n_pt = cur_pt - num_cols;
	 s_pt = cur_pt + num_cols;

	 w_w_pt = w_pt - 1;
	 e_e_pt = e_pt + 1;
	 n_n_pt = n_pt - num_cols;
	 s_s_pt = s_pt + num_cols;

	 nw_pt = n_pt - 1;
	 ne_pt = n_pt + 1;
	 sw_pt = s_pt - 1;
	 se_pt = s_pt + 1;

	 sceVu0FVECTOR up = {0, 0, 0, 0}, down = {0, 0, 0, 0}, left = {0, 0, 0, 0}, right = {0, 0, 0, 0};

	 /*
	 vec_xyzw up(0, 0, 0, 0), down, left, right;
	 down = left = right = up;
	 */

	 // stretch forces

	 float spring_const = 4.0f;
	 float damp_const = 1.5f;
	 float rest_length_row = stretch_rest_length_row,
	    rest_length_col = stretch_rest_length_col;

	 if ( row != 0 ) 
	    get_spring_force_lv( up, *cur_pt, *n_pt, spring_const, damp_const, rest_length_col );
	 if ( row != (num_rows-1) ) 
	    get_spring_force_lv( down, *cur_pt, *s_pt, spring_const, damp_const, rest_length_col );
	 if ( col != 0 )
	    get_spring_force_lv( left, *cur_pt, *w_pt, spring_const, damp_const, rest_length_row );
	 if ( col != (num_cols-1) )
	    get_spring_force_lv( right, *cur_pt, *e_pt, spring_const, damp_const, rest_length_row );

	 // shear forces

	 spring_const = 4.0f;
	 damp_const = 1.5f;
	 float rest_length = shear_rest_length;

	 sceVu0FVECTOR top_bslash = {0, 0, 0, 0}, top_fslash = {0, 0, 0, 0};
	 sceVu0FVECTOR bottom_bslash = {0, 0, 0, 0}, bottom_fslash = {0, 0, 0, 0};
	 /*
	 vec_xyzw top_bslash(0, 0, 0, 0), top_fslash, bottom_fslash, bottom_bslash;
	 top_fslash = bottom_bslash = bottom_fslash = top_bslash;
	 */

	 if ( row != 0 ) {
	    if ( col != 0 )
	       get_spring_force_lv( top_bslash, *cur_pt, *nw_pt, spring_const, damp_const, rest_length );
	    if ( col != (num_cols-1) )
	       get_spring_force_lv( top_fslash, *cur_pt, *ne_pt, spring_const, damp_const, rest_length );
	 }
	 if ( row != (num_rows-1) ) {
	    if ( col != 0 )
	       get_spring_force_lv( bottom_fslash, *cur_pt, *sw_pt, spring_const, damp_const, rest_length );
	    if ( col != (num_cols-1) )
	       get_spring_force_lv( bottom_bslash, *cur_pt, *se_pt, spring_const, damp_const, rest_length );
	 }

	 // bend forces

	 spring_const = 3.0f;
	 damp_const = 1.4f;
	 rest_length_row = bend_rest_length_row;
	 rest_length_col = bend_rest_length_col;

	 sceVu0FVECTOR n_bend = {0, 0, 0, 0}, s_bend = {0, 0, 0, 0};
	 sceVu0FVECTOR e_bend = {0, 0, 0, 0}, w_bend = {0, 0, 0, 0};
	 /*
	 vec_xyzw n_bend(0, 0, 0, 0), s_bend, e_bend, w_bend;
	 s_bend = e_bend = w_bend = n_bend;
	 */

	 if ( row > 1 )
	    get_spring_force_lv( n_bend, *cur_pt, *n_n_pt, spring_const, damp_const, rest_length_col );
	 if ( row < (num_rows-2) )
	    get_spring_force_lv( s_bend, *cur_pt, *s_s_pt, spring_const, damp_const, rest_length_col );
	 if ( col > 1 )
	    get_spring_force_lv( w_bend, *cur_pt, *w_w_pt, spring_const, damp_const, rest_length_row );
	 if ( col < (num_cols-2) )
	    get_spring_force_lv( e_bend, *cur_pt, *e_e_pt, spring_const, damp_const, rest_length_row );

	 sceVu0FVECTOR force;

	 sceVu0CopyVector( force, up );
	 sceVu0AddVector( force, force, down );
	 sceVu0AddVector( force, force, left );
	 sceVu0AddVector( force, force, right );

	 sceVu0AddVector( force, force, top_fslash );
	 sceVu0AddVector( force, force, top_bslash );
	 sceVu0AddVector( force, force, bottom_fslash );
	 sceVu0AddVector( force, force, bottom_bslash );

	 sceVu0AddVector( force, force, n_bend );
	 sceVu0AddVector( force, force, s_bend );
	 sceVu0AddVector( force, force, e_bend );
	 sceVu0AddVector( force, force, w_bend );

	 /*
	 vec_xyzw force = up;
	 force += down;
	 force += left;
	 force += right;

	 force += top_fslash;
	 force += top_bslash;
	 force += bottom_bslash;
	 force += bottom_fslash;

	 force += n_bend;
	 force += s_bend;
	 force += e_bend;
	 force += w_bend;
	 */

	 // atmospheric drag

	 sceVu0FVECTOR temp;
	 sceVu0ScaleVector( temp, cur_pt->velocity, -0.3f );
	 sceVu0AddVector( force, force, temp );
	 /*
	 force += cur_pt->velocity * -0.3f;
	 */

	 // wind

	 float temp_f = sceVu0InnerProduct( cur_pt->normal, wind_dir );
	 sceVu0ScaleVector( temp, cur_pt->normal, temp_f * wind_strength );
	 sceVu0AddVector( force, force, temp );
	 /*
	 vec_xyzw force_dir = cur_pt->normal;
	 force += force_dir * ((float)cur_pt->normal.dot(wind_dir) * wind_strength);
	 */

	 // gravity

	 sceVu0AddVector( force, force, gravity_lv );
	 sceVu0FVECTOR delta_v, delta_pos;
	 sceVu0ScaleVector( delta_v, force, timestep );
	 sceVu0ScaleVector( delta_pos, cur_pt->velocity, timestep );
	 /*
	 force += gravity;
					
	 vec_xyzw accel = force; // assuming mass of 1
	 vec_xyzw delta_v = accel * timestep;
	 vec_xyzw delta_pos = cur_pt->velocity * timestep;
	 */

	 // now fixup the lengths of the springs so that they (mostly) don't exceed
	 // their rest lengths by 10%.  Seems ok to only do this for the stretch
	 // springs

	 float length;
	 sceVu0FVECTOR direction, new_pos;
	 sceVu0AddVector( new_pos, cur_pt->position, delta_pos );
	 /*
	 vec_xyzw direction;
			
	 vec_xyzw new_pos = cur_pt->position + delta_pos;
	 */

	 if ( row != 0 ) {
	    sceVu0SubVector( direction, n_pt->position, new_pos );
	    length = sqrtf( sceVu0InnerProduct(direction, direction) );
	    if ( length > stretch_rest_length_col * 1.1f ) {
	       sceVu0ScaleVector( direction, direction, 1.0f / length );
	       sceVu0ScaleVector( direction,
				  direction, length - stretch_rest_length_col * 1.1f );
	       sceVu0AddVector( new_pos, new_pos, direction );
	    }
	    else if ( length < stretch_rest_length_col * 0.9f ) {
	       sceVu0ScaleVector( direction, direction, 1.0f / length );
	       sceVu0ScaleVector( direction,
				  direction, length - stretch_rest_length_col * 0.9f );
	       sceVu0AddVector( new_pos, new_pos, direction );
	    }
	 }

	 if ( row != (num_rows-1) ) {
	    sceVu0SubVector( direction, s_pt->position, new_pos );
	    length = sqrtf( sceVu0InnerProduct(direction, direction) );
	    if ( length > stretch_rest_length_col * 1.1f ) {
	       sceVu0ScaleVector( direction, direction, 1.0f / length );
	       sceVu0ScaleVector( direction,
				  direction, length - stretch_rest_length_col * 1.1f );
	       sceVu0AddVector( new_pos, new_pos, direction );
	    }
	    else if ( length < stretch_rest_length_col * 0.9f ) {
	       sceVu0ScaleVector( direction, direction, 1.0f / length );
	       sceVu0ScaleVector( direction,
				  direction, length - stretch_rest_length_col * 0.9f );
	       sceVu0AddVector( new_pos, new_pos, direction );
	    }
	 }

	 if ( col != (num_cols-1) ) {
	    sceVu0SubVector( direction, e_pt->position, new_pos );
	    length = sqrtf( sceVu0InnerProduct(direction, direction) );
	    if ( length > stretch_rest_length_row * 1.1f ) {
	       sceVu0ScaleVector( direction, direction, 1.0f / length );
	       sceVu0ScaleVector( direction,
				  direction, length - stretch_rest_length_row * 1.1f );
	       sceVu0AddVector( new_pos, new_pos, direction );
	    }
	    else if ( length < stretch_rest_length_row * 0.9f ) {
	       sceVu0ScaleVector( direction, direction, 1.0f / length );
	       sceVu0ScaleVector( direction,
				  direction, length - stretch_rest_length_row * 0.9f );
	       sceVu0AddVector( new_pos, new_pos, direction );
	    }
	 }

	 if ( col != 0 ) {
	    sceVu0SubVector( direction, w_pt->position, new_pos );
	    length = sqrtf( sceVu0InnerProduct(direction, direction) );
	    if ( length > stretch_rest_length_row * 1.1f ) {
	       sceVu0ScaleVector( direction, direction, 1.0f / length );
	       sceVu0ScaleVector( direction,
				  direction, length - stretch_rest_length_row * 1.1f );
	       sceVu0AddVector( new_pos, new_pos, direction );
	    }
	    else if ( length < stretch_rest_length_row * 0.9f ) {
	       sceVu0ScaleVector( direction, direction, 1.0f / length );
	       sceVu0ScaleVector( direction,
				  direction, length - stretch_rest_length_row * 0.9f );
	       sceVu0AddVector( new_pos, new_pos, direction );
	    }
	 }

	 // find the velocity that would have resulted in the new final position
	 sceVu0SubVector( delta_pos, new_pos, cur_pt->position );
	 sceVu0FVECTOR new_vel;
	 sceVu0ScaleVector( new_vel, delta_pos, 1.0f / timestep );
	 /*
	 delta_pos = new_pos - cur_pt->position;
	 vec_xyzw new_vel = delta_pos / timestep;
	 */

	 // fix the leftmost col in place
	 if ( col == 0 ) {
	    *new_pt = *cur_pt;
	 }
	 else {
	    sceVu0CopyVector( new_pt->position, new_pos );
	    sceVu0AddVector( cur_pt->velocity, new_vel, delta_v );
	    sceVu0AddVector( new_pt->velocity, new_vel, delta_v );
	    /*
	    new_pt->position = new_pos;
	    // this has a nice stabilizing effect on the system.. I think the bend
	    // springs are losing it when the fabric becomes folded...
	    cur_pt->velocity = new_vel + delta_v;
	    new_pt->velocity = new_vel + delta_v;
	    */
	 }

	 new_pt++;
	 cur_pt++;
      }
   }

   update_normals( (cloth_pt*)new_pts );

   // copy new points back
   Utils::MemCpy128( (tU128*)points, (tU128*)new_pts, num_cols*num_rows * (sizeof(cloth_pt_libvu0)/16) );

   mStopTimer("cloth::update_libvu0");
}

/********************************************
 * end
 */

void
cloth::draw( CVifSCDmaPacket& packet, const mat_44& tex_xform )
{
   mStartTimer("cloth::draw");

   int col, row;
   vec_xyzw tex_uv(0, 0, 1, 0);

   // start at row 1
   cloth_pt *cur_pt = points + num_cols, *up_pt;

   packet.Pad128();
   for ( row = 1; row < num_rows; row++ )
   {
      // xfer the vertices, normals, and tex uvs

      packet.Pad96();
      packet.OpenUnpack( Vifs::UnpackModes::v4_32, kInputStart, Packet::kDoubleBuff );
      {
	 for ( col = 0; col < num_cols; col++ ) {
	    up_pt = cur_pt - num_cols;

	    packet += up_pt->position;
	    packet += up_pt->normal;
	    tex_uv = vec_xy(tex_xform * up_pt->position);
	    packet += tex_uv;

	    packet += cur_pt->position;
	    packet += cur_pt->normal;
	    tex_uv = vec_xy(tex_xform * cur_pt->position);
	    packet += tex_uv;

	    cur_pt++;
	 }
      }
      packet.CloseUnpack();

      // xfer header info

      xfer_buffer_header( packet, 3, num_cols*2, 0, NULL );
   }

   mStopTimer("cloth::draw");
}

void
cloth::xfer_buffer_header( CVifSCDmaPacket &packet,
			   int vu1QuadsPerVert,
			   int numVerts, int numStrips,
			   unsigned short *stripOffsets )
{
   packet.Stcycl(4,4);
   packet.Pad96();
   packet.OpenUnpack( Vifs::UnpackModes::v4_32, 0, Packet::kDoubleBuff );
   {
      // num verts
      packet += (tU64)numVerts; packet += (tU64)0;

      // strip offsets
      unsigned int adcBits = 0xc000;
      unsigned int adc, stopBit = 0x2000;
      for ( int i = 0; i < 16; i++ ) {
	 if ( i < numStrips - 1 )
	    adc = (unsigned int)(adcBits | stripOffsets[i] * vu1QuadsPerVert );
	 else if ( i == numStrips - 1 )
	    adc = (unsigned int)(adcBits | stopBit | stripOffsets[i] * vu1QuadsPerVert );
	 else
	    adc = (unsigned int)0;
	 packet += adc;
      }
   }
   packet.CloseUnpack();

   packet.Mscnt();
   packet.Pad128();
}

void
cloth::print()
{
   int i;
   cloth_pt *pt = points;
   for ( i = 0; i < 10; i++, pt++ )
      pt->position.print();
}

void
cloth::update_normals( cloth_pt* pts ) {
   cloth_pt *cur_pt = pts;

   vec_xyzw up, down, left, right;
   vec_xyzw normal;
   cloth_pt *right_pt, *left_pt, *up_pt, *down_pt;

   int col, row;

   // ul corner
   right_pt = cur_pt + 1;
   down_pt = cur_pt + num_cols;
   right = right_pt->position - cur_pt->position;
   down = down_pt->position - cur_pt->position;
   normal = down.cross(right);
   normal.normalize();
   cur_pt->normal = normal;
   cur_pt++;

   // do top middle
   for ( col = 1; col < (num_cols-1); col++ ) {
      left_pt = cur_pt - 1;
      right_pt = cur_pt + 1;
      down_pt = cur_pt + num_cols;

      left = left_pt->position - cur_pt->position;
      right = right_pt->position - cur_pt->position;
      down = down_pt->position - cur_pt->position;

      normal = left.cross(down);
      normal += down.cross(right);

      cur_pt->normal = normal.normalize();
      cur_pt++;
   }

   // ur corner
   left_pt = cur_pt - 1;
   down_pt = cur_pt + num_cols;
   left = left_pt->position - cur_pt->position;
   down = down_pt->position - cur_pt->position;
   normal = left.cross(down);
   normal.normalize();
   cur_pt->normal = normal;
   cur_pt++;

   // middle section
   for ( row = 1; row < (num_rows-1); row++ ) {
      up_pt = cur_pt - num_cols;
      down_pt = cur_pt + num_cols;
      right_pt = cur_pt + 1;

      up = up_pt->position - cur_pt->position;
      down = down_pt->position - cur_pt->position;
      right = right_pt->position - cur_pt->position;

      normal = down.cross(right);
      normal += right.cross(up);
      cur_pt->normal = normal.normalize();
      cur_pt++;
				
      for ( col = 1; col < (num_cols-1); col++ ) {
	 left_pt = cur_pt - 1;
	 right_pt = cur_pt + 1;
	 up_pt = cur_pt - num_cols;
	 down_pt = cur_pt + num_cols;

	 up = up_pt->position - cur_pt->position;
	 down = down_pt->position - cur_pt->position;
	 left = left_pt->position - cur_pt->position;
	 right = right_pt->position - cur_pt->position;

	 normal = right.cross(up);
	 normal += up.cross(left);
	 normal += left.cross(down);
	 normal += down.cross(right);
	 normal.normalize();

	 cur_pt->normal = normal;
	 cur_pt++;
      }

      up_pt = cur_pt - num_cols;
      down_pt = cur_pt + num_cols;
      left_pt = cur_pt - 1;

      up = up_pt->position - cur_pt->position;
      down = down_pt->position - cur_pt->position;
      left = left_pt->position - cur_pt->position;

      normal = up.cross(left);
      normal += left.cross(down);
      cur_pt->normal = normal.normalize();
      cur_pt++;
   }			

   // ll corner
   right_pt = cur_pt + 1;
   up_pt = cur_pt - num_cols;
   right = right_pt->position - cur_pt->position;
   up = up_pt->position - cur_pt->position;
   normal = right.cross(up);
   cur_pt->normal = normal.normalize();;
   cur_pt++;

   // bottom middle
   for ( col = 1; col < (num_cols-1); col++ ) {
      left_pt = cur_pt - 1;
      right_pt = cur_pt + 1;
      up_pt = cur_pt - num_cols;

      left = left_pt->position - cur_pt->position;
      right = right_pt->position - cur_pt->position;
      up = up_pt->position - cur_pt->position;

      normal = up.cross(left);
      normal += right.cross(up);

      cur_pt->normal = normal.normalize();
      cur_pt++;
   }

   // lr corner
   left_pt = cur_pt - 1;
   up_pt = cur_pt - num_cols;
   left = left_pt->position - cur_pt->position;
   up = up_pt->position - cur_pt->position;
   normal = up.cross(left);
   cur_pt->normal = normal.normalize();
   cur_pt++;
}
